from fastapi import APIRouter, Depends, Response, Request
from starlette_context import context
import os
import logging
from requests.structures import CaseInsensitiveDict
from auth import auth
import requests
import conf

import _version as b
import correlation
import socket
import k8s
import time
import psutil
# If doing tracing enable import pdb

from pydantic import BaseModel

from starlette.status import (
    HTTP_400_BAD_REQUEST,
)

router = APIRouter()


class Detail(BaseModel):
    detail: str


def seconds_elapsed():
    return time.time() - psutil.boot_time()


@router.get(
    "/config",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
    },
)
def show_policy_config_details(
    response: Response,
    request: Request,
    auth_data: auth.Auth = Depends(auth.require_authorized_admin),
):
    """
    Return detail configuration details.
    Diagnostic API

    ### Permission required to use this API:

        policy.service.admin
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    # if doing tracing insert pdb.set_trace()
    r_status_json = {}
    r_config_json = {}
    r_config_headers = {}
    correlation_id = context["correlation_id"]
    user_agent = context["user_agent"]
    headers = CaseInsensitiveDict()
    headers["X-Request-ID"] = correlation_id
    headers["X-Correlation-ID"] = correlation_id
    headers["Correlation-ID"] = correlation_id
    headers["req_id"] = correlation_id

    configmap = k8s.get_opa_config(namespace=conf.NAMESPACE, name=conf.OPA_CONFIG_MAP)
    try:
        status_url = conf.OPA_STATUS_API
        r_status = requests.get(status_url, timeout=10, headers=headers)
        r_status_json = r_status.json()
    except requests.exceptions.HTTPError:
        logger.error(f"Error: endpoint {status_url}: HTTPError")
    except Exception:
        logger.error(f"Error: endpoint {status_url}: Error")

    try:
        r_config = requests.get(conf.OPA_CONFIG_API, timeout=10, headers=headers)
        r_config_json = r_config.json()
        r_config_headers = r_config.headers
    except requests.exceptions.HTTPError:
        logger.error(f"Error: endpoint {conf.OPA_CONFIG_API}: HTTPError")
    except Exception:
        logger.error(f"Error: endpoint {conf.OPA_CONFIG_API}: Error")

    cpu_usage = get_cpu_usage()
    memory_usage = get_memory_usage()

    conf_dict = {}
    for i in dir(conf):
        if i[0].isupper():
            if "MOCK_ENTITLEMENT_RESULT" not in i and "TOKEN" not in i:
                conf_dict[i] = getattr(conf, i)

    if conf.MASK_OPA_CONFIG_DETAILS:
        r_config_json = None

    if conf.MASK_CONFIG_ENV:
        env = {
            "CLOUD_PROVIDER": os.environ.get("CLOUD_PROVIDER"),
            "HOSTNAME": os.environ.get("HOSTNAME"),
            "PYTHON_VERSION": os.environ.get("PYTHON_VERSION"),
            "KUBERNETES_SERVICE_HOST": os.environ.get("KUBERNETES_SERVICE_HOST"),
        }
    else:
        env = os.environ.items()

    logger.info(f"config [{user_agent} {auth_data.x_user_id}]")

    hostname = socket.gethostname()
    uptime = seconds_elapsed()
    return {
        "request": {"correlation_id": correlation_id, "user_agent": user_agent},
        "opa_config_response_headers": r_config_headers,
        "conf": conf_dict,
        "configmap": configmap,
        "env": env,
        "hostname": hostname,
        "uptime": uptime,
        "cpu_usage": cpu_usage,
        "memory_usage": memory_usage,
        "client_host": request.client.host,
        "request_headers": request.headers,
        "auth_data": {
            "data_partition_id": auth_data.data_partition_id,
        },
        "build_data": {
            "version": b.__version__,
            "milestone": b.__milestone__,
            "artifactId": None,
            "name": "policy",
            "groupId": "org.opengroup.osdu",
            "buildTime": b.__committimestamp__,
            "branch": b.__branch__,
            "commitId": b.__commitid__,
            "commitMessage": b.__commitmessage__,
            "commitrefslug": b.__commitrefslug__,
        },
        "opa_status": r_status_json,
        "opa_config": r_config_json,
    }


def get_cpu_usage():
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    try:
        with open("/sys/fs/cgroup/cpu/cpuacct.usage") as file:
            return file.read().strip()
    except FileNotFoundError as err:
        logger.error(f"file not found cpu_usage error: {err}")
    except OSError as err:
        logger.error(f"cpu_usage error: {err}")

    return None


def get_memory_usage():
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    try:
        with open("/sys/fs/cgroup/memory/memory.usage_in_bytes") as file:
            return file.read().strip()
    except FileNotFoundError as err:
        logger.error(f"file not found memory_usage error: {err}")
    except OSError as err:
        logger.error(f"memory_usage error: {err}")
