from fastapi import APIRouter, HTTPException
import requests
import logging
from pydantic import BaseModel

from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_501_NOT_IMPLEMENTED,
    HTTP_503_SERVICE_UNAVAILABLE,
)

import conf


class Detail(BaseModel):
    detail: str


router = APIRouter()


@router.get(
    "/health",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_503_SERVICE_UNAVAILABLE: {"model": Detail},
    },
)
async def health():
    """
    ## Health check endpoint, which does not depend on OPA.
    This API does not require any headers or authentication.

    The /health endpoint responds with a 200 HTTP status code when the service pod can receive requests.
    The endpoint indicates that the service pod is healthy and reachable.
    It does not indicate that the service is ready to serve requests.

    ### Permission required to use this API:

        None
    """
    return {"message": "Healthy"}


@router.get(
    "/ready",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_501_NOT_IMPLEMENTED: {"model": Detail},
        HTTP_503_SERVICE_UNAVAILABLE: {"model": Detail},
    },
)
def ready():
    """
    ## Health check endpoint, which depends on OPA being available and healthy.
    This API does not require any headers or authentication.

    ### Possible http return status codes:
    * 200 - no error
    * 501 - not implemented
    * 503 - service not available

    The /ready endpoint responds with a 200 HTTP status code if the overall application works.
    The endpoint indicates that the service is ready to serve requests.

    ### Permission required to use this API:

        None
    """
    try:
        url = conf.OPA_HEALTH_API
    except NameError:
        logging.error("conf.OPA_URL undefined")
        raise HTTPException(
            status_code=HTTP_501_NOT_IMPLEMENTED,
            detail="NotHealthy - OPA configuration issue",
        )
    except AttributeError:
        logging.error("conf.OPA_URL undefined")
        raise HTTPException(
            status_code=HTTP_501_NOT_IMPLEMENTED,
            detail="NotHealthy - OPA configuration issue",
        )

    try:
        r = requests.get(url, timeout=10)
        r.raise_for_status()

    except requests.exceptions.HTTPError as e:
        logging.error(f"Http Error to OPA endpoint: {url}: {e}")
        raise HTTPException(
            status_code=HTTP_503_SERVICE_UNAVAILABLE,
            detail="NotHealthy - OPA HTTPError",
        )
    except requests.exceptions.ConnectionError as e:
        logging.error(f"Error Connecting to OPA endpoint: {url}: {e}")
        raise HTTPException(
            status_code=HTTP_503_SERVICE_UNAVAILABLE,
            detail="NotHealthy - OPA ConnectionError",
        )
    except requests.exceptions.Timeout as e:
        logging.error(f"Timeout Error to OPA endpoint: {url}: {e}")
        raise HTTPException(
            status_code=HTTP_503_SERVICE_UNAVAILABLE, detail="NotHealthy - OPA Timeout"
        )
    except requests.exceptions.RequestException as e:
        logging.error(f"Connection error to OPA endpoint: {url}: {e}")
        raise HTTPException(
            status_code=HTTP_503_SERVICE_UNAVAILABLE, detail="NotHealthy - OPA Error"
        )
    return {"message": "Ready"}
