import fastapi
from fastapi import Depends, HTTPException, Response
from auth import auth
import logging
import conf
from starlette_context import context
import k8s
import correlation
import ruamel.yaml
from kubernetes.client.models import V1ConfigMap, V1ObjectMeta
from pydantic import BaseModel

from starlette.status import (
    HTTP_202_ACCEPTED,
    HTTP_400_BAD_REQUEST,
    HTTP_405_METHOD_NOT_ALLOWED,
    HTTP_404_NOT_FOUND,
    HTTP_501_NOT_IMPLEMENTED,
    HTTP_502_BAD_GATEWAY,
)

router = fastapi.APIRouter()


class Detail(BaseModel):
    detail: str


@router.get(
    "/tenant",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_405_METHOD_NOT_ALLOWED: {"model": Detail},
        HTTP_501_NOT_IMPLEMENTED: {"model": Detail},
    },
)
def get_tenant(
    response: Response,
    all_data: bool = False,
    auth_data: auth.Auth = Depends(auth.require_authorized_admin),
):
    """
    Experimental tenant API for retrieving OPA bundle config for a data partition.
    These details are read from OPA configmap.

    ### Permission required to use this API:

    `policy.service.admin`
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    yaml = ruamel.yaml.YAML()
    config_map = k8s.get_opa_config(namespace=conf.NAMESPACE, name=conf.OPA_CONFIG_MAP)
    if config_map:
        data = yaml.load(config_map)
        path = f"osdu/partition/{auth_data.data_partition_id}"
        if all_data:
            return config_map
        elif path in config_map:
            ret_data = {path: data["bundles"][path]}
            return ret_data
        else:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND,
                detail=f"{path} not found in OPA configmap.",
            )
    else:
        logger.info("GET /tenant HTTP_405_METHOD_NOT_ALLOWED")
        raise HTTPException(
            status_code=HTTP_405_METHOD_NOT_ALLOWED,
            detail=f"Unable to read configmap '{conf.OPA_CONFIG_MAP}' details",
        )


def _mlput(self, path, value, index=None, sep=":"):
    spath = path.split(sep)
    parent = self.mlget(spath[:-1])
    if index is None:
        parent[spath[-1]] = value
    else:
        parent.insert(index, spath[-1], value)


update_tenant_description = f"""
Experimental tenant API for updating OPA bundle config for a data partition.

Adding new partitions is not supported in M20 and later.

Services must be one of the supported (s3, gcs, gcp, blob, ...). See documentation for full list.
- Min polling 1 second
- Max polling 3600(1 hour)

### Permission required to use this API on this system (configurable):

{conf.OPS_PERMISSION}
"""


@router.put(
    "/tenant",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_405_METHOD_NOT_ALLOWED: {"model": Detail},
        HTTP_501_NOT_IMPLEMENTED: {"model": Detail},
        HTTP_502_BAD_GATEWAY: {"model": Detail},
    },
    description=update_tenant_description,
)
def update_tenant(
    response: Response,
    service: str,
    polling_min_delay_seconds: int = 6,
    polling_max_delay_seconds: int = 12,
    auth_data: auth.Auth = Depends(auth.require_authorized_ops),
):
    """
    Experimental tenant API for updating OPA bundle config for a data partition.
    Adding new partitions is not supported in M20.

    Services must be one of the supported (s3, gcs, gcp, blob, ...). See documentation for full list.
    Min polling 1 second
    Max polling 3600(1 hour)
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    logger.info(
        f"service: {service}, polling_min_delay_seconds: {polling_min_delay_seconds}, polling_max_delay_seconds: {polling_max_delay_seconds}"
    )

    if service not in conf.OPA_SUPPORTED_SERVICES:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="Unsupported service provided"
        )

    if (
        polling_min_delay_seconds >= polling_max_delay_seconds
        or polling_min_delay_seconds < 1
        or polling_max_delay_seconds > conf.OPA_MAX_POLLING
    ):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Unsupported polling_min_delay_seconds polling_max_delay_seconds",
        )

    config_map = k8s.get_opa_config(
        namespace=conf.NAMESPACE, name=conf.OPA_CONFIG_MAP, config_only=False
    )
    if config_map:
        path = f"osdu/partition/{auth_data.data_partition_id}"
        yaml = ruamel.yaml.YAML(typ=["rt", "string"])
        ruamel.yaml.comments.CommentedMap.mlput = _mlput

        print(type(config_map.data))
        for key, val in config_map.data.items():
            logger.info(f"path: {path} key: {key} config_map: {config_map}")
            logger.info(f"key: {key} type: {type(val)} val: {val}")

            data = yaml.load(config_map.data[key])
            if path in config_map.data[key]:
                logger.info(f"path found: {data['bundles'][path]}")
                data["bundles"][path]["service"] = service
                data["bundles"][path][
                    "resource"
                ] = f"bundle-{auth_data.data_partition_id}.tar.gz"
                if "polling" in data["bundles"][path]:
                    data["bundles"][path]["polling"][
                        "min_delay_seconds"
                    ] = polling_min_delay_seconds
                    data["bundles"][path]["polling"][
                        "max_delay_seconds"
                    ] = polling_max_delay_seconds
                else:
                    logger.warning("polling not found in bundle configuration")

                new_data_str = yaml.dump_to_string(data)
                logger.debug(f"new_data_str: {new_data_str}")

                body = V1ConfigMap(
                    metadata=V1ObjectMeta(
                        name=conf.OPA_CONFIG_MAP,
                        labels={
                            "app": "opa",
                            "app.kubernetes.io/managed-by": "policy-service",
                        },
                    ),
                    kind="ConfigMap",
                    data={key: new_data_str},
                )

                res, msg = k8s.patch_opa_config(
                    namespace=conf.NAMESPACE, name=conf.OPA_CONFIG_MAP, body=body
                )
                if res:
                    raise HTTPException(
                        status_code=HTTP_202_ACCEPTED, detail=f"Accepted {res}"
                    )
                else:
                    raise HTTPException(
                        status_code=HTTP_502_BAD_GATEWAY,
                        detail=f"An error occurred while talking to kubernetes {msg}",
                    )

            else:
                raise HTTPException(
                    status_code=HTTP_501_NOT_IMPLEMENTED,
                    detail=f"{path} not found in {data}",
                )

    else:
        logger.info("PUT /tenant HTTP_405_METHOD_NOT_ALLOWED")
        raise HTTPException(
            status_code=HTTP_405_METHOD_NOT_ALLOWED,
            detail="Unable to read configmap details",
        )


delete_tenant_description = f"""
Experimental tenant API for deleting tenant OPA bundle config for a data partition.

- Deleting partitions is not supported in M21.
- X-Error response header is also included.

### Permission required to use this API on this system (configurable):

{conf.OPS_PERMISSION}
"""


@router.delete(
    "/tenant",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_405_METHOD_NOT_ALLOWED: {"model": Detail},
        HTTP_501_NOT_IMPLEMENTED: {"model": Detail},
    },
    description=delete_tenant_description,
)
def delete_tenant(
    response: Response,
    auth_data: auth.Auth = Depends(auth.require_authorized_ops),
):
    """
    Experimental tenant API for deleting tenant OPA bundle config for a data partition.
    Deleting partitions is not supported in M21.
    X-Error response header is also included.
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    config_map = k8s.get_opa_config(namespace=conf.NAMESPACE, name=conf.OPA_CONFIG_MAP)
    if config_map:
        yaml = ruamel.yaml.YAML()
        data = yaml.load(config_map)
        path = f"osdu/partition/{auth_data.data_partition_id}"
        del data["bundles"][path]
        raise HTTPException(
            status_code=HTTP_501_NOT_IMPLEMENTED,
            detail=f"Not Implemented for {path}",
            headers={"X-Error": data},
        )
    else:
        logger.info("DELETE /tenant HTTP_405_METHOD_NOT_ALLOWED")
        raise HTTPException(
            status_code=HTTP_405_METHOD_NOT_ALLOWED,
            detail="Unable to read configmap details",
            headers={"X-Error": "prohibited_by_environment"},
        )
