import conf
import requests
import logging
import json
from starlette_context import context
from opa_response import OpaResponse
import correlation
from cachetools import cached, TTLCache
import requests.exceptions

# import httpx
import timer

from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

logger = logging.getLogger(__name__)


# cache policy data for no longer than TTL seconds
@cached(
    cache=TTLCache(maxsize=conf.OPA_LIST_CACHE_MAXSIZE, ttl=conf.OPA_LIST_CACHE_TTL)
)
def fetch_opa_policies_direct(timeout=20):
    """fetch a list of all policies in OPA"""
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)

    result = OpaResponse()
    try:
        rsp = requests.get(conf.OPA_POLICIES_API, timeout=timeout)
        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        if rsp.ok:
            # logging.info(f"result from OPA ({conf.OPA_POLICIES_API}): {json.dumps(result.json, indent=2)}")
            logger.debug(f"result from OPA ({conf.OPA_POLICIES_API}) OK")
        else:
            logger.error(
                "Error fetching policies, got status %s and message: %s"
                % (rsp.status_code, rsp.text)
            )
            if "message" in result.json:
                result.message = result.json["message"]
            elif "code" in result.json:
                result.message = result.json["code"]
    except Exception as err:
        logger.error(f"Error fetching policies from OPA: {err}")
    return result


def fetch_opa_policy_direct(policy_id, timeout=20, name="fetch_opa_policy"):
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    t = timer.Timer(name, logger=logger.debug)
    t.start()
    result = fetch_opa_policy_direct_with_caching(policy_id=policy_id, timeout=timeout)
    t.stop()
    return result


@cached(
    cache=TTLCache(maxsize=conf.OPA_FETCH_CACHE_MAXSIZE, ttl=conf.OPA_FETCH_CACHE_TTL)
)
def fetch_opa_policy_direct_with_caching(policy_id, timeout=20):
    """
    fetch a specific policy in OPA
    200 - no error
    500 - server error
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    url = conf.OPA_POLICIES_API + "/" + policy_id

    logger.debug(f"OPA get URL: {url}")
    result = OpaResponse()
    try:
        rsp = requests.get(url, timeout=timeout)
        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        if rsp.ok:
            # logging.debug(json.dumps(result.json, indent=2))
            logger.debug(f"get {url} OK")
        else:
            logger.error(
                f"Error fetching policies from {url}, got status {rsp.status_code} and message: {rsp.text}"
            )
            if "message" in result.json:
                result.message = result.json["message"]
            elif "code" in result.json:
                result.message = result.json["code"]

    except Exception as err:
        logger.error(err)
    return result


def delete_opa_policy_direct(policy_id, timeout=20):
    """
    Delete a policy directly from OPA
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    url = conf.OPA_POLICIES_API + "/" + policy_id
    result = OpaResponse()
    result.message = f"Error while removing policy {policy_id}"

    try:
        rsp = requests.delete(url, timeout=timeout)
        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        if rsp.ok:
            logger.info(
                f"deleted policy {policy_id} from {url}, and message: {rsp.text}"
            )
            logger.debug(json.dumps(result.json, indent=2))
            if result.json == {}:
                result.message = f"Policy {policy_id} removed from OPA server"
            else:
                result.message = f"Policy {policy_id} removed from OPA server, but expected return {result.json}"
        else:
            logger.error(
                f"Error deleting policy {policy_id} from {url}, got status {rsp.status_code} and message: {rsp.text}"
            )
            if "message" in result.json:
                result.message = result.json["message"]
            elif "code" in result.json:
                result.message = result.json["code"]

    except Exception as err:
        logger.error(err)
    return result


def put_opa_policy_direct(policy_id, data, timeout=20, headers=None, validate=False):
    """
    Put a policy in OPA directly for use when not using a CLOUD_PROVIDER
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    url = conf.OPA_POLICIES_API + "/" + policy_id
    logger.info(f"OPA put URL {url}")
    result = OpaResponse(status_code=503)

    if not headers:
        headers = headers = {"Content-Type": "application/octet-stream"}

    try:
        rsp = requests.put(url, data=data, headers=headers, timeout=timeout)
        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        if rsp.ok:
            logger.debug(
                f"Response from OPA({url}): {json.dumps(result.json, indent=2)}"
            )
            result.message = f"Policy {policy_id} added to OPA"
        else:
            if validate:
                logger.info(
                    "Error updating policy for validation OPA direct, got status %s and message: %s"
                    % (rsp.status_code, rsp.text)
                )
            else:
                logger.error(
                    "Error updating policy with OPA direct, got status %s and message: %s"
                    % (rsp.status_code, rsp.text)
                )
            result.message = f"{rsp.text} {rsp.status_code}"
    except requests.ConnectionError as err:
        logger.error(
            f"Unexpected ConnectionError when attempting to connect to OPA: {err}"
        )
        result.message = "Unexpected ConnectionError when connecting to OPA"
    except Exception as err:
        logger.error(f"Unexpected error when attempting to connect to OPA: {err}")
        result.message = "Unexpected error when connecting to OPA"
    return result


def compile(query, metrics=False, instrument=False, timeout=20, name="compile api"):
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    t = timer.Timer(name, logger=logger.debug)
    t.start()
    result = compile_with_caching(
        query=query, metrics=metrics, instrument=instrument, timeout=timeout
    )
    t.stop()
    return result


@cached(
    cache=TTLCache(
        maxsize=conf.OPA_COMPILE_CACHE_MAXSIZE, ttl=conf.OPA_COMPILE_CACHE_TTL
    )
)
def compile_with_caching(query, metrics=False, instrument=False, timeout=20):
    """
    Compile - Partially Evaluate a Query.
    The Compile API allows you to partially evaluate Rego queries and obtain a simplified version of the policy.

    OPA can report detailed performance metrics at runtime.
    Performance metrics can be requested on individual API calls and are returned inline with the API response.
    To enable performance metric collection on an API call, specify the metrics=true query parameter when executing the API call.

    OPA also supports query instrumentation.
    To enable query instrumentation, specify the instrument=true query parameter when executing the API call.
    Query instrumentation can help diagnose performance problems, however, it can add significant overhead to query evaluation.
    We recommend leaving query instrumentation off unless you are debugging a performance problem.
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)

    url = conf.OPA_COMPILE_API
    if metrics:
        url = url + "?metrics=true"

    if instrument and metrics:
        url = url + "&instrument=true"

    logger.debug(f"OPA compile URL {url} type: {type(query)} json_data: {query}")
    result = OpaResponse()

    try:
        rsp = requests.post(
            url,
            data=query,
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )

        result.status_code = rsp.status_code
        result.ok = rsp.ok
        result.json = rsp.json()
        result.text = rsp.text

        if rsp.ok:
            logger.debug(
                f"OK Response from OPA({url}): {json.dumps(result.json, indent=2)}"
            )
            result.message = "Policy OPA compiled"
        else:
            logger.error(
                "Error compiling policy, got status %s and message: %s"
                % (rsp.status_code, rsp.text)
            )
            result.message = f"Error compiling policy: {rsp.text}"

    except Exception as err:
        logger.error(err)
        try:
            result.message = (
                f"Error compiling policy: {query}, code: {rsp.status_code}, {rsp.text}"
            )
            result.status_code = rsp.status_code
        except NameError:
            result.message = f"Error compiling policy: {query}"
            result.status_code = HTTP_422_UNPROCESSABLE_ENTITY
        result.ok = False
    return result


def data(query, path, timeout=20, name="opa data api"):
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    t = timer.Timer(name, logger=logger.debug)
    t.start()
    result = data_with_caching(query=query, path=path, timeout=timeout)
    t.stop()
    return result


@cached(
    cache=TTLCache(
        maxsize=conf.OPA_DATAAPI_CACHE_MAXSIZE, ttl=conf.OPA_DATAAPI_CACHE_TTL
    )
)
def data_with_caching(query, path, timeout=20):
    """
    Compile - Partially Evaluate a Query.
    The Compile API allows you to partially evaluate Rego queries and obtain a simplified version of the policy.
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    url = conf.OPA_DATA_API + "/" + path
    logger.debug(f"OPA compile URL {url} type: {type(query)} json_data: {query}")
    result = OpaResponse()

    try:
        # async with httpx.AsyncClient() as client:
        rsp = requests.post(
            url,
            # rsp = await client.post(url,
            data=query,
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )

        result.status_code = rsp.status_code
        # result.ok = (rsp.status_code == httpx.codes.OK)
        result.ok = rsp.ok
        result.json = rsp.json()
        result.text = rsp.text

        # if rsp.status_code == httpx.codes.OK:
        if rsp.ok:
            logger.debug(
                f"OK Response from OPA({url}): {json.dumps(result.json, indent=2)}"
            )
            result.message = "Policy OPA query"
        else:
            logger.error(
                "Error compiling policy, got status %s and message: %s"
                % (rsp.status_code, rsp.text)
            )
            result.message = f"Error query policy: {rsp.text}"

    except Exception as err:
        logger.error(err)
    return result


def get_document_with_input(rego_package_name, rego_rule_name, rego_input, timeout=20):
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    t = timer.Timer("get document with input", logger=logger.debug)
    t.start()
    result = get_document_with_input_cached(
        rego_package_name=rego_package_name,
        rego_rule_name=rego_rule_name,
        rego_input=rego_input,
        timeout=timeout,
    )
    t.stop()
    return result


@cached(
    cache=TTLCache(
        maxsize=conf.OPA_DATAAPI_CACHE_MAXSIZE, ttl=conf.OPA_DATAAPI_CACHE_TTL
    )
)
def get_document_with_input_cached(
    rego_package_name, rego_rule_name, rego_input, timeout=20
):
    """
    Get data from OPA
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    url = conf.OPA_DATA_API + "/" + rego_package_name + "/" + rego_rule_name
    logger.debug(f"get_document_with_input url: {url}")
    try:
        # async with httpx.AsyncClient() as client:
        # response = await client.post(url,
        resp = requests.post(
            url, data=json.dumps({"input": rego_input}), timeout=timeout
        )

    except Exception as err:
        logger.error(err)

    result = resp.json()["result"]
    return result
